
#ifndef __TOOLS_H_INCLUDED
#define __TOOLS_H_INCLUDED

#include <vector>
#include <string>

/// \a for loop replacement - counts from 0 to length-1
/// \param index Name of the index variable to declare
/// \param length Number of loop iterations
/// \remarks This \a for loop replacement counts from \a 0 to \a length-1
#define FOR_INDEX( index, length )\
	for ( size_t index=0; index < length; index++ )

/// \a for loop replacement - iterates over a collection
/// \param it Name of iterator to declare
/// \param collection Collection variable or reference for which to iterate
/// \param type Collection type
/// \remarks The collection type is required for declaring the iterator variable
#define FOR_ITER( it, collection, type )\
	for ( type::iterator it=(collection).begin(); it != (collection).end(); it++ )

/// \a for loop replacement - iterates over a const collection
/// \param it Name of const_iterator to declare
/// \param collection Collection variable or reference for which to iterate
/// \param type Collection type
/// \remarks The collection type is required for declaring the iterator variable
#define FOR_CONST_ITER( it, collection, type )\
	for ( type::const_iterator it=(collection).begin(); it != (collection).end(); it++ )

/// \a for loop replacement - iterates over a collection in a reverse order
/// \param it Name of reverse_iterator to declare
/// \param collection Collection variable or reference for which to iterate
/// \param type Collection type
/// \remarks The collection type is required for declaring the iterator variable
#define FOR_REV_ITER( it, collection, type )\
	for ( type::reverse_iterator it=(collection).rbegin(); it != (collection).rend(); it++ )

/// \a for loop replacement - iterates over a const collection in a reverse order
/// \param it Name of const_reverse_iterator to declare
/// \param collection Collection variable or reference for which to iterate
/// \param type Collection type
/// \remarks The collection type is required for declaring the iterator variable
#define FOR_CONST_REV_ITER( it, collection, type )\
	for ( type::const_reverse_iterator it=(collection).rbegin(); it != (collection).rend(); it++ )

/// Converts \a int to \a string
/// \param value \a int value to convert
/// \return Value converted to user-readable string
std::string Int2String( int value );

/// Fills a variable with zeroes
/// \param p Pointer to a variable that has only bare variables
template< class T >
inline void ZeroFill( T* p )
{
    memset( p, 0, sizeof(T) );
}

/// Converts a binary 32-bit float representation to a float
inline float FloatFromU32( unsigned i )
{
	return *reinterpret_cast<float*>( &i );
}

#endif // __TOOLS_H_INCLUDED

// vim:ts=4:sts=4:et
