
#include "main_window.h"
#include "resource.h"
#include "d3d_pixel_shader.h"
#include "../Common/tools.h"

using namespace WinAPI;

namespace {
    Application< MainWindow > app;

    WindowClass main_wc(
        /* hIcon        */  LoadIcon( __MODULE_HANDLE, MAKEINTRESOURCE(IDI_SHADERCONVD3D) ),
        /* hbrBackground*/  reinterpret_cast<HBRUSH>( COLOR_3DFACE+1 )
        );

    Font sysfont( "MS Shell Dlg", -MulDiv( 8, GetDeviceCaps(ClientDC(GetDesktopWindow()),LOGPIXELSY), 72 ) );
    Font textfont( "Courier New", -MulDiv( 10, GetDeviceCaps(ClientDC(GetDesktopWindow()),LOGPIXELSY), 72 ) );

    enum {
        ID_COMPILE = 201,
        ID_ALPHABLEND,
    };
}

MainWindow::MainWindow()
{
    INITCOMMONCONTROLSEX icc = { sizeof( INITCOMMONCONTROLSEX ), ICC_WIN95_CLASSES | ICC_COOL_CLASSES };
    InitCommonControlsEx( &icc );

    TEXTMETRIC tm;
    {   ClientDC dc( GetDesktopWindow() );
        SelectObj sel( dc, sysfont );
        dc.GetTextMetrics( &tm );
    }
    x_dlu = tm.tmAveCharWidth;
    y_dlu = tm.tmHeight + tm.tmAscent;
    btn_width = MulDiv( 50, x_dlu, 4 );
    btn_height = MulDiv( 8, y_dlu, 8 );

    window_width = 768 + 2*GetSystemMetrics( SM_CXBORDER );
    window_height = 256 + btn_height + 2*GetSystemMetrics( SM_CYBORDER ) + GetSystemMetrics( SM_CYCAPTION ) + MulDiv( 2, y_dlu, 8 );

    Create( "ShaderConvD3D", main_wc, WS_OVERLAPPEDWINDOW, WRect(WPoint(CW_USEDEFAULT,0),window_width,window_height) );

    // Create D3D object
    p_d3d = IPtr<IDirect3D9>( Direct3DCreate9( D3D_SDK_VERSION ) );

    // Get caps and determine vp flags
    p_d3d->GetDeviceCaps( D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, &caps );
    DWORD behavior = D3DCREATE_SOFTWARE_VERTEXPROCESSING;
    if ( caps.DevCaps & D3DDEVCAPS_HWTRANSFORMANDLIGHT ) {
        behavior = D3DCREATE_HARDWARE_VERTEXPROCESSING;
        if ( caps.DevCaps & D3DDEVCAPS_PUREDEVICE )
            behavior |= D3DCREATE_PUREDEVICE;
    }

    // Get current adapter mode
    D3DDISPLAYMODE mode;
    p_d3d->GetAdapterDisplayMode( D3DADAPTER_DEFAULT, &mode );

    // Prepare presentation params
    D3DPRESENT_PARAMETERS params;
    ZeroMemory( &params, sizeof(params) );
    params.BackBufferWidth = 256;
    params.BackBufferHeight = 256;
    params.BackBufferFormat = mode.Format;
    params.SwapEffect = D3DSWAPEFFECT_COPY;
    params.Windowed = TRUE;
    params.EnableAutoDepthStencil = TRUE;
    params.AutoDepthStencilFormat = D3DFMT_D16;

    // Create device
    const HRESULT hr = p_d3d->CreateDevice( D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, stc_image, behavior, &params, p_device.GetPP() );
    if ( FAILED(hr) ) {
        p_d3d.Release();
        char errorstr[64];
        sprintf( errorstr, "%08X", hr );
        MessageBox( errorstr, "Direct3D Error", MB_OK | MB_ICONSTOP );
        b_created = false;
        Destroy();
        return;
    }

    // Create textures
    p_texture.resize( caps.MaxTextureBlendStages );
    for ( DWORD i=0; i < caps.MaxSimultaneousTextures; i++ )
		D3DXCreateTextureFromFile( p_device, ("tex"+Int2String(i)+".png").c_str(), p_texture[i].GetPP() );
    D3DXCreateTextureFromFile( p_device, "background.png", p_background.GetPP() );

    // Set up filtering
    for ( DWORD i=0; i < caps.MaxTextureBlendStages; i++ )
        SetFiltering( i );

    // Initiate drawing
    QueryPerformanceCounter( &base_time );
    SetTimer( *this, 0, 100, 0 );
}

void MainWindow::SetFiltering( DWORD stage ) const
{
    if ( caps.TextureFilterCaps & D3DPTFILTERCAPS_MAGFANISOTROPIC )
        p_device->SetSamplerState( stage, D3DSAMP_MAGFILTER, D3DTEXF_ANISOTROPIC );
    else if ( caps.TextureFilterCaps & D3DPTFILTERCAPS_MAGFLINEAR )
        p_device->SetSamplerState( stage, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR );

    if ( caps.TextureFilterCaps & D3DPTFILTERCAPS_MINFANISOTROPIC )
        p_device->SetSamplerState( stage, D3DSAMP_MINFILTER, D3DTEXF_ANISOTROPIC );
    else if ( caps.TextureFilterCaps & D3DPTFILTERCAPS_MINFLINEAR )
        p_device->SetSamplerState( stage, D3DSAMP_MINFILTER, D3DTEXF_LINEAR );

    if ( caps.TextureFilterCaps & D3DPTFILTERCAPS_MIPFLINEAR )
        p_device->SetSamplerState( stage, D3DSAMP_MIPFILTER, D3DTEXF_LINEAR );
    else if ( caps.TextureFilterCaps & D3DPTFILTERCAPS_MIPFPOINT )
        p_device->SetSamplerState( stage, D3DSAMP_MIPFILTER, D3DTEXF_POINT );

    if ( caps.MaxAnisotropy > 1 )
        p_device->SetSamplerState( stage, D3DSAMP_MAXANISOTROPY, caps.MaxAnisotropy );
}

bool MainWindow::OnCreate( LPCREATESTRUCT lpCreateStruct )
{
    stc_image.Create( "", WRect(0,0,256,256), *this, -1, SS_BLACKRECT );
    ed_shader.Create( "", WRect(256,0,512,256), *this, -1, ES_LEFT|ES_AUTOHSCROLL|ES_AUTOVSCROLL|WS_HSCROLL|WS_VSCROLL|ES_MULTILINE );
    ed_code.Create( "", WRect(512,0,768,256), *this, -1, ES_LEFT|ES_AUTOHSCROLL|ES_AUTOVSCROLL|WS_HSCROLL|WS_VSCROLL|ES_MULTILINE|ES_READONLY );
    btn_compile.Create( "Compile", WRect(0,0,1,1), *this, ID_COMPILE );
    chk_alphablend.Create( "Alpha Blending", WRect(x_dlu,257,x_dlu+btn_width*2,256+btn_height), *this, ID_ALPHABLEND );
    ed_shader.SetFont( textfont );
    ed_code.SetFont( textfont );
    btn_compile.SetFont( sysfont );
    chk_alphablend.SetFont( sysfont );
    return true;
}

void MainWindow::OnDestroy()
{
    KillTimer( *this, 0 );
    p_background.Release();
    p_texture.clear();
    p_device.Release();
    p_d3d.Release();
    PostQuitMessage( 0 );
}

int MainWindow::HandleMessage( int uMsg, int wParam, int lParam )
{
    switch ( uMsg ) {

    case WM_GETMINMAXINFO:
        {   LPMINMAXINFO p_mmi = windows_cast<LPMINMAXINFO>( lParam );
            p_mmi->ptMinTrackSize = WPoint( window_width, window_height );
            p_mmi->ptMaxTrackSize.y = window_height;
            p_mmi->ptMaxSize.y = window_height;
        }
        return 0;

    case WM_TIMER:
        DrawFrame();
        return 0;

    default:
        return Window::HandleMessage( uMsg, wParam, lParam );

    }
}

namespace {

    void ConvertEOLs( string& s )
    {
        for ( size_t i=0; i < s.length(); i++ )
            if ( s[i] == 10 && i > 0 && s[i-1] != 13 )
                s.insert( i, 1, '\r' );
    }

}

void MainWindow::OnCommand( int nIdentifier, int nNotifyCode, HWND hwndControl )
{
    switch ( nIdentifier ) {

    case ID_COMPILE:
        {   // Compile pixel shader to binary code
            ed_code.SetText( "" );
            const string source = ed_shader.GetText();
            IPtr<ID3DXBuffer> p_shader, p_errors;
            if ( FAILED( D3DXAssembleShader( source.data(), int(source.length()), 0, 0, D3DXSHADER_DEBUG, p_shader.GetPP(), p_errors.GetPP() ) ) ) {
                string errors( (const char*)p_errors->GetBufferPointer() );
                ConvertEOLs( errors );
                ed_code.SetText( errors.c_str() );
                return;
            }

            // Transform binary code into instructions
            vector<INSTRUCTION> instructions;
            try {
                vector<unsigned> shader_code( p_shader->GetBufferSize()/4 );
                const char* src = (const char*)p_shader->GetBufferPointer();
                for ( size_t i=0; i < shader_code.size(); i++, src += 4 )
                    shader_code[i] = *reinterpret_cast<const unsigned*>( src );
                DisasmShader( shader_code, &instructions );
            } catch ( ShaderError error ) {
                string errors = "Disassemble error\r\n";
                if ( error.LineNumber() >= 0 )
                    errors += "(line " + Int2String(error.LineNumber()) + ") ";
                errors += error.ErrorMsg();
                ed_code.SetText( errors.c_str() );
                return;
            }

            // Display code
            string s = GetReadableShader( instructions );
            ConvertEOLs( s );
            ed_code.SetText( s.c_str() );

            // Transform instructions into TSS settings
            vector<DWORD> tss;
            try {
                FEATUREMAP features( caps );
                BuildStateList( instructions, features, &tss, false );
            } catch ( ShaderError error ) {
                string errors = "Direct3D TSS error\r\n";
                if ( error.LineNumber() >= 0 )
                    errors += "(line " + Int2String(error.LineNumber()) + ") ";
                errors += error.ErrorMsg();
                ed_code.SetText( errors.c_str() );
                return;
            }

            // Check TSS
            SetTSS( tss );
            DWORD num_passes = 0;
            const bool b_ok = SUCCEEDED( p_device->ValidateDevice( &num_passes ) );
            DefaultTSS();
            if ( ! b_ok || num_passes != 1 ) {
                ed_code.SetText( "Device validation failed" );
                return;
            }
            stage_states.swap( tss );

            // Display states
            s += GetReadableTSS( stage_states );
            ConvertEOLs( s );
            ed_code.SetText( s.c_str() );
        }
        break;

    default:
        Window::OnCommand( nIdentifier, nNotifyCode, hwndControl );

    }
}

void MainWindow::OnSize( int sizing, WinAPI::WSize new_size )
{
    const int width = (new_size.cx - 256) >> 1;
    ed_code.SetPos( WRect(new_size.cx-width,0,new_size.cx,256) );
    ed_shader.SetPos( WRect(256,0,new_size.cx-width,256) );
    WRect btn_rc( new_size.cx-btn_width, 256, new_size.cx, 256+btn_height );
    btn_compile.SetPos( btn_rc );
}

void MainWindow::DrawFrame()
{
    p_device->Clear( 0, 0, D3DCLEAR_TARGET, 0, 1, 0 );
    p_device->BeginScene();

    p_device->SetRenderState( D3DRS_ZENABLE, FALSE );
    p_device->SetRenderState( D3DRS_ZWRITEENABLE, FALSE );

    if ( p_background.IsValid() ) {
        D3DVIEWPORT9 viewport;
        p_device->GetViewport( &viewport );
        const float x1 = viewport.X-0.5f;
        const float y1 = viewport.Y-0.5f;
        const float x2 = viewport.X+viewport.Width-0.5f;
        const float y2 = viewport.Y+viewport.Height-0.5f;

        struct VERTEX {
            float x, y, z, w;
            float u, v;
        };
        VERTEX vertices[] = {
            { x1, y2, 1, 1, 0, 1 },
            { x1, y1, 1, 1, 0, 0 },
            { x2, y2, 1, 1, 1, 1 },
            { x2, y1, 1, 1, 1, 0 },
        };

        p_device->SetTexture( 0, p_background );
        p_device->SetFVF( D3DFVF_XYZRHW | D3DFVF_TEX1 );
        p_device->DrawPrimitiveUP( D3DPT_TRIANGLESTRIP, 2, vertices, sizeof(VERTEX) );
    }

    D3DXMATRIX matrix;
    D3DXMatrixPerspectiveFovLH( &matrix, D3DX_PI*45.0f/180.0f, 1, 1, 100 );
    p_device->SetTransform( D3DTS_PROJECTION, &matrix );
    D3DXMatrixLookAtLH( &matrix, &D3DXVECTOR3(0,0,-3), &D3DXVECTOR3(0,0,0), &D3DXVECTOR3(0,1,0) );
    p_device->SetTransform( D3DTS_VIEW, &matrix );

    LARGE_INTEGER time, frequency;
    QueryPerformanceCounter( &time );
    QueryPerformanceFrequency( &frequency );
    const float timef = float( (time.QuadPart-base_time.QuadPart) * 1000 / frequency.QuadPart );
    D3DXMatrixRotationY( &matrix, D3DX_PI*45.0f/180.0f*timef/1000.f );
    p_device->SetTransform( D3DTS_WORLD, &matrix );

    D3DLIGHT9 light;
    ZeroMemory( &light, sizeof(light) );
    light.Type = D3DLIGHT_POINT;
    light.Diffuse = light.Ambient = light.Specular = D3DXCOLOR( 1, 0.2f, 0.2f, 1 );
    light.Position = D3DXVECTOR3( -3, 3, -30 );
    light.Range = 1000;
    light.Attenuation0 = 1;
    p_device->SetLight( 0, &light );
    p_device->LightEnable( 0, TRUE );
    light.Diffuse = light.Ambient = light.Specular = D3DXCOLOR( 0.2f, 1, 1, 1 );
    light.Position = D3DXVECTOR3( 3, -3, -30 );
    p_device->SetLight( 1, &light );
    p_device->LightEnable( 1, TRUE );
    p_device->SetRenderState( D3DRS_LIGHTING, TRUE );
    p_device->SetRenderState( D3DRS_AMBIENT, 0xFF404040 );

    D3DMATERIAL9 mat;
    mat.Ambient = D3DXCOLOR( 0.2f, 0.2f, 0.2f, 1 );
    mat.Diffuse = D3DXCOLOR( 1, 1, 1, 1 );
    mat.Emissive = D3DXCOLOR( 0, 0, 0, 0 );
    mat.Specular = D3DXCOLOR( 0, 0, 0, 0 );
    mat.Power = 0;
    p_device->SetMaterial( &mat );

    for ( DWORD i=0; i < caps.MaxTextureBlendStages; i++ )
        if ( p_texture[i].IsValid() )
            p_device->SetTexture( i, p_texture[i] ),
            p_device->SetTextureStageState( i, D3DTSS_TEXCOORDINDEX, 0 );

    if ( ! stage_states.empty() )
        SetTSS( stage_states );

    if ( IsDlgButtonChecked( ID_ALPHABLEND ) == BST_CHECKED )
        p_device->SetRenderState( D3DRS_ALPHABLENDENABLE, TRUE ),
        p_device->SetRenderState( D3DRS_SRCBLEND, D3DBLEND_SRCALPHA ),
        p_device->SetRenderState( D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA );
    else
        p_device->SetRenderState( D3DRS_ALPHABLENDENABLE, FALSE );

    struct VERTEX {
        float x, y, z;
        float nx, ny, nz;
        float u, v;
    };
    VERTEX vertices[] = {
        {-1, 1, 0, 0, 0,-1, 0, 0 },
        { 1, 1, 0, 0, 0,-1, 1, 0 },
        { 1,-1, 0, 0, 0,-1, 1, 1 },

        {-1, 1, 0, 0, 0,-1, 0, 0 },
        { 1,-1, 0, 0, 0,-1, 1, 1 },
        {-1,-1, 0, 0, 0,-1, 0, 1 },

        {-1, 1, 0, 0, 0, 1, 0, 0 },
        { 1,-1, 0, 0, 0, 1, 1, 1 },
        { 1, 1, 0, 0, 0, 1, 1, 0 },

        {-1, 1, 0, 0, 0, 1, 0, 0 },
        {-1,-1, 0, 0, 0, 1, 0, 1 },
        { 1,-1, 0, 0, 0, 1, 1, 1 },
    };

    p_device->SetFVF( D3DFVF_XYZ | D3DFVF_NORMAL | D3DFVF_TEX1 );
    p_device->DrawPrimitiveUP( D3DPT_TRIANGLELIST, 4, vertices, sizeof(VERTEX) );

    for ( DWORD i=0; i < caps.MaxTextureBlendStages; i++ )
        p_device->SetTexture( i, 0 );
    DefaultTSS();
    p_device->SetRenderState( D3DRS_ALPHABLENDENABLE, FALSE );

    p_device->EndScene();
    p_device->Present( 0, 0, 0, 0 );
}

void MainWindow::SetTSS( const vector<DWORD>& states ) const
{
    DWORD stage = 0;
    for ( size_t i=0; i < states.size()/2; i++ ) {
        const DWORD reg = states[i*2];
        const DWORD value = states[i*2+1];
        if ( reg == STAGESELECT ) {
            stage = value;
            continue;
        }
        if ( reg == D3DRS_TEXTUREFACTOR ) {
            p_device->SetRenderState( D3DRS_TEXTUREFACTOR, value );
            continue;
        }
        p_device->SetTextureStageState( stage, D3DTEXTURESTAGESTATETYPE(reg), value );
    }
}

void MainWindow::DefaultTSS() const
{
    for ( DWORD stage=0; stage < caps.MaxTextureBlendStages; stage++ ) {
        p_device->SetTextureStageState( stage, D3DTSS_COLOROP, D3DTOP_DISABLE );
        p_device->SetTextureStageState( stage, D3DTSS_COLORARG1, D3DTA_TEXTURE );
        p_device->SetTextureStageState( stage, D3DTSS_COLORARG2, D3DTA_CURRENT );
        p_device->SetTextureStageState( stage, D3DTSS_ALPHAOP, D3DTOP_DISABLE );
        p_device->SetTextureStageState( stage, D3DTSS_ALPHAARG1, D3DTA_TEXTURE );
        p_device->SetTextureStageState( stage, D3DTSS_ALPHAARG2, D3DTA_CURRENT );
        p_device->SetTextureStageState( stage, D3DTSS_RESULTARG, D3DTA_CURRENT );
    }
    p_device->SetTextureStageState( 0, D3DTSS_COLOROP, D3DTOP_MODULATE );
    p_device->SetTextureStageState( 0, D3DTSS_ALPHAOP, D3DTOP_SELECTARG1 );
}

// vim:ts=4:sts=4:et

