
#include "main_window.h"
#include "resource.h"
#include "../Common/tools.h"
#include "../PNGDecoder/png_decoder.h"
#include <GL/glu.h>

using namespace WinAPI;

namespace {
    Application< MainWindow > app;

    WindowClass main_wc(
        /* hIcon        */  LoadIcon( __MODULE_HANDLE, MAKEINTRESOURCE(IDI_SHADERCONVGL) ),
        /* hbrBackground*/  reinterpret_cast<HBRUSH>( COLOR_3DFACE+1 )
        );

    Font sysfont( "MS Shell Dlg", -MulDiv( 8, GetDeviceCaps(ClientDC(GetDesktopWindow()),LOGPIXELSY), 72 ) );
    Font textfont( "Courier New", -MulDiv( 10, GetDeviceCaps(ClientDC(GetDesktopWindow()),LOGPIXELSY), 72 ) );

    enum {
        ID_COMPILE = 201,
        ID_ALPHABLEND,
    };
}

MainWindow::MainWindow()
: hdc( 0 ), hrc( 0 ), background( 0 )
{
    INITCOMMONCONTROLSEX icc = { sizeof( INITCOMMONCONTROLSEX ), ICC_WIN95_CLASSES | ICC_COOL_CLASSES };
    InitCommonControlsEx( &icc );

    TEXTMETRIC tm;
    {   ClientDC dc( GetDesktopWindow() );
        SelectObj sel( dc, sysfont );
        dc.GetTextMetrics( &tm );
    }
    x_dlu = tm.tmAveCharWidth;
    y_dlu = tm.tmHeight + tm.tmAscent;
    btn_width = MulDiv( 50, x_dlu, 4 );
    btn_height = MulDiv( 8, y_dlu, 8 );

    window_width = 768 + 2*GetSystemMetrics( SM_CXBORDER );
    window_height = 256 + btn_height + 2*GetSystemMetrics( SM_CYBORDER ) + GetSystemMetrics( SM_CYCAPTION ) + MulDiv( 2, y_dlu, 8 );

    Create( "ShaderConvGL", main_wc, WS_OVERLAPPEDWINDOW, WRect(WPoint(CW_USEDEFAULT,0),window_width,window_height) );

    hdc = GetDC( stc_image );

    // Set pixel format
    PIXELFORMATDESCRIPTOR pfd;
    ZeroMemory( &pfd, sizeof(pfd) );
    pfd.nSize       = sizeof(pfd);
    pfd.nVersion    = 1;
    pfd.dwFlags     = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
    pfd.iPixelType  = PFD_TYPE_RGBA;
    pfd.cColorBits  = 24;
    pfd.cDepthBits  = 16;
    pfd.iLayerType  = PFD_MAIN_PLANE;
    const int i_format = ChoosePixelFormat( hdc, &pfd );
    SetPixelFormat( hdc, i_format, &pfd );

    // Create OpenGL rendering context
    hrc = wglCreateContext( hdc );
    const bool b_all_ok = hrc && wglMakeCurrent( hdc, hrc );

    // Destroy window if anything went wrong
    if ( ! b_all_ok ) {
        b_created = false;
        Destroy();
        return;
    }

    // Get extension functions
    glActiveTexture = (PFNGLACTIVETEXTUREARBPROC)wglGetProcAddress( "glActiveTextureARB" );
    glClientActiveTexture = (PFNGLCLIENTACTIVETEXTUREARBPROC)wglGetProcAddress( "glClientActiveTextureARB" );
    glMultiTexCoord2f = (PFNGLMULTITEXCOORD2FARBPROC)wglGetProcAddress( "glMultiTexCoord2fARB" );
    GLFEATUREMAP features;
    max_texture_units = features.max_texture_units;

    // Load textures
    background = LoadTexture( "background.png" );
    texture.resize( max_texture_units, 0 );
    for ( int i=0; i < max_texture_units; i++ )
        texture[i] = LoadTexture( "tex"+Int2String(i)+".png" );

    // Initiate drawing
    QueryPerformanceCounter( &base_time );
    SetTimer( *this, 0, 100, 0 );
}

bool MainWindow::OnCreate( LPCREATESTRUCT lpCreateStruct )
{
    stc_image.Create( "", WRect(0,0,256,256), *this, -1, SS_BLACKRECT );
    ed_shader.Create( "", WRect(256,0,512,256), *this, -1, ES_LEFT|ES_AUTOHSCROLL|ES_AUTOVSCROLL|WS_HSCROLL|WS_VSCROLL|ES_MULTILINE );
    ed_code.Create( "", WRect(512,0,768,256), *this, -1, ES_LEFT|ES_AUTOHSCROLL|ES_AUTOVSCROLL|WS_HSCROLL|WS_VSCROLL|ES_MULTILINE|ES_READONLY );
    btn_compile.Create( "Compile", WRect(0,0,1,1), *this, ID_COMPILE );
    chk_alphablend.Create( "Alpha Blending", WRect(x_dlu,257,x_dlu+btn_width*2,256+btn_height), *this, ID_ALPHABLEND );
    ed_shader.SetFont( textfont );
    ed_code.SetFont( textfont );
    btn_compile.SetFont( sysfont );
    chk_alphablend.SetFont( sysfont );
    return true;
}

void MainWindow::OnClose()
{
    if ( background ) glDeleteTextures( 1, &background ), background = 0;
    for ( size_t i=0; i < texture.size(); i++ )
        if ( texture[i] ) glDeleteTextures( 1, &texture[i] );
    texture.clear();
    wglMakeCurrent( 0, 0 );
    try {
        if ( hrc ) wglDeleteContext( hrc ), hrc = 0;
    } catch (...) {
        MessageBox( "Weird error!", "Warning", MB_OK | MB_ICONWARNING );
    }
    if ( hdc ) ReleaseDC( stc_image, hdc ), hdc = 0;
    Destroy();
}

void MainWindow::OnDestroy()
{
    KillTimer( *this, 0 );
    PostQuitMessage( 0 );
}

int MainWindow::HandleMessage( int uMsg, int wParam, int lParam )
{
    switch ( uMsg ) {

    case WM_GETMINMAXINFO:
        {   LPMINMAXINFO p_mmi = windows_cast<LPMINMAXINFO>( lParam );
            p_mmi->ptMinTrackSize = WPoint( window_width, window_height );
            p_mmi->ptMaxTrackSize.y = window_height;
            p_mmi->ptMaxSize.y = window_height;
        }
        return 0;

    case WM_TIMER:
        DrawFrame();
        return 0;

    default:
        return Window::HandleMessage( uMsg, wParam, lParam );

    }
}

namespace {

    void ConvertEOLs( string& s )
    {
        for ( size_t i=0; i < s.length(); i++ )
            if ( s[i] == 10 && i > 0 && s[i-1] != 13 )
                s.insert( i, 1, '\r' );
    }

}

std::string AssembleShader( const std::string& source_code, std::vector<unsigned int>* p_executable );

void MainWindow::OnCommand( int nIdentifier, int nNotifyCode, HWND hwndControl )
{
    switch ( nIdentifier ) {

    case ID_COMPILE:
        {   // Compile pixel shader to binary code
            ed_code.SetText( "" );
            const string source = ed_shader.GetText();
            vector<unsigned> shader_code;
            {   string errors = AssembleShader( source, &shader_code );
                if ( ! errors.empty() ) {
                    ConvertEOLs( errors );
                    ed_code.SetText( errors.c_str() );
                    return;
                }
            }

            // Transform binary code into instructions
            vector<INSTRUCTION> instructions;
            try {
                DisasmShader( shader_code, &instructions );
            } catch ( ShaderError error ) {
                string errors = "Disassemble error\r\n";
                if ( error.LineNumber() >= 0 )
                    errors += "(line " + Int2String(error.LineNumber()) + ") ";
                errors += error.ErrorMsg();
                ed_code.SetText( errors.c_str() );
                return;
            }

            // Display code
            string s = GetReadableShader( instructions );
            ConvertEOLs( s );
            ed_code.SetText( s.c_str() );

            // Transform instructions into TexEnv settings
            vector<TEXENVVALUE> tss;
            try {
                GLFEATUREMAP features;
                BuildTexEnvSetup( instructions, features, &tss, true );
            } catch ( ShaderError error ) {
                string errors = "OpenGL TexEnv setup error\r\n";
                if ( error.LineNumber() >= 0 )
                    errors += "(line " + Int2String(error.LineNumber()) + ") ";
                errors += error.ErrorMsg();
                ed_code.SetText( errors.c_str() );
                return;
            }

            // Check TexEnv setup
            const bool b_ok = SetTexEnv( tss );
            DefaultTexEnv();
            if ( ! b_ok ) {
                ed_code.SetText( "Device validation failed" );
                return;
            }
            texenv_setup.swap( tss );

            // Display states
            s += GetReadableTexEnv( texenv_setup );
            ConvertEOLs( s );
            ed_code.SetText( s.c_str() );
        }
        break;

    default:
        Window::OnCommand( nIdentifier, nNotifyCode, hwndControl );

    }
}

void MainWindow::OnSize( int sizing, WinAPI::WSize new_size )
{
    const int width = (new_size.cx - 256) >> 1;
    ed_code.SetPos( WRect(new_size.cx-width,0,new_size.cx,256) );
    ed_shader.SetPos( WRect(256,0,new_size.cx-width,256) );
    WRect btn_rc( new_size.cx-btn_width, 256, new_size.cx, 256+btn_height );
    btn_compile.SetPos( btn_rc );
}

namespace {

    struct GLCOLOR {
        GLfloat r, g, b, a;

        GLCOLOR( GLfloat r, GLfloat g, GLfloat b, GLfloat a = 1 ): r( r ), g( g ), b( b ), a( a ) { }
    };

    struct GLPOINT {
        GLfloat x, y, z, w;

        GLPOINT( GLfloat x, GLfloat y, GLfloat z, GLfloat w = 1 ): x( x ), y( y ), z( z ), w( w ) { }
    };

    struct GLVECTOR {
        GLfloat x, y, z;

        GLVECTOR( GLfloat x, GLfloat y, GLfloat z ): x( x ), y( y ), z( z ) { }
    };

    typedef GLfloat GLSCALAR;

    struct GLMATERIAL {
        GLCOLOR ambient;
        GLCOLOR diffuse;
        GLCOLOR specular;
        GLCOLOR emissive;
        GLSCALAR power;

        GLMATERIAL();
        void Set( GLenum face = GL_FRONT);
    };

    GLMATERIAL::GLMATERIAL()
    :   ambient( 0.2f, 0.2f, 0.2f ),
        diffuse( 0.8f, 0.8f, 0.8f ),
        specular( 0.0f, 0.0f, 0.0f ),
        emissive( 0.0f, 0.0f, 0.0f ),
        power( 0.0f )
    {
    }

    void GLMATERIAL::Set( GLenum face )
    {
        glMaterialfv( face, GL_AMBIENT, &ambient.r );
        glMaterialfv( face, GL_DIFFUSE, &diffuse.r );
        glMaterialfv( face, GL_SPECULAR, &specular.r );
        glMaterialfv( face, GL_EMISSION, &emissive.r );
        glMaterialf( face, GL_SHININESS, power );
    }

    struct GLLIGHT {
        GLCOLOR ambient;
        GLCOLOR diffuse;
        GLCOLOR specular;
        GLPOINT position;
        GLVECTOR direction;
        GLSCALAR exponent;
        GLSCALAR cutoff;
        GLSCALAR attenuation[3];

        GLLIGHT();
        void Set( GLenum light );
    };

    GLLIGHT::GLLIGHT()
    :   ambient( 0.0f, 0.0f, 0.0f ),
        diffuse( 1.0f, 1.0f, 1.0f ),
        specular( 1.0f, 1.0f, 1.0f ),
        position( 0.0f, 0.0f, 1.0f, 0.0f ),
        direction( 0.0f, 0.0f, -1.0f ),
        exponent( 0.0f ),
        cutoff( 180.0f )
    {
        attenuation[0] = 1.0f;
        attenuation[1] = 0.0f;
        attenuation[2] = 0.0f;
    }

    void GLLIGHT::Set( GLenum light )
    {
        glLightfv( light, GL_AMBIENT, &ambient.r );
        glLightfv( light, GL_DIFFUSE, &diffuse.r );
        glLightfv( light, GL_SPECULAR, &specular.r );
        glLightfv( light, GL_POSITION, &position.x );
        glLightfv( light, GL_SPOT_DIRECTION, &direction.x );
        glLightf( light, GL_SPOT_EXPONENT, exponent );
        glLightf( light, GL_SPOT_CUTOFF, cutoff );
        glLightf( light, GL_CONSTANT_ATTENUATION, attenuation[0] );
        glLightf( light, GL_LINEAR_ATTENUATION, attenuation[1] );
        glLightf( light, GL_QUADRATIC_ATTENUATION, attenuation[2] );
    }

}

void MainWindow::DrawFrame()
{
    glClear( GL_COLOR_BUFFER_BIT );

    glEnable( GL_CULL_FACE );

    if ( background ) {
        glDisable( GL_LIGHTING );

        glMatrixMode( GL_MODELVIEW );
        glLoadIdentity();

        glMatrixMode( GL_PROJECTION );
        glLoadIdentity();
        glOrtho( 0, 256, 0, 256, 1, 1000 );

        glEnable( GL_TEXTURE_2D );
        glBindTexture( GL_TEXTURE_2D, background );
        glTexEnvf( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE );

        glColor3f( 1, 1, 1 );
        glBegin( GL_TRIANGLE_STRIP );
            glTexCoord2f( 0, 0 );
            glVertex3f( 0, 0, -10 );
            glTexCoord2f( 1, 0 );
            glVertex3f( 256, 0, -10 );
            glTexCoord2f( 0, 1 );
            glVertex3f( 0, 256, -10 );
            glTexCoord2f( 1, 1 );
            glVertex3f( 256, 256, -10 );
        glEnd();
    }

    glMatrixMode( GL_PROJECTION );
    glLoadIdentity();
    gluPerspective( 45, 1, 1, 100 );

    glMatrixMode( GL_MODELVIEW );
    glLoadIdentity();
    gluLookAt( 0, 0, 3, 0, 0, 0, 0, 1, 0 );

    glEnable( GL_LIGHTING );
    glLightModelfv( GL_LIGHT_MODEL_AMBIENT, &(GLCOLOR( 0.25f, 0.25f, 0.25f ).r) );

    GLLIGHT light;
    light.ambient = light.diffuse = light.specular = GLCOLOR( 1, 0.2f, 0.2f );
    light.position = GLPOINT( -3, 3, 30 );
    light.Set( GL_LIGHT0 );
    glEnable( GL_LIGHT0 );

    light.ambient = light.diffuse = light.specular = GLCOLOR( 0.2f, 1, 1 );
    light.position = GLPOINT( 3, -3, 30 );
    light.Set( GL_LIGHT1 );
    glEnable( GL_LIGHT1 );

    GLMATERIAL mat;
    mat.diffuse = GLCOLOR( 1, 1, 1 );
    mat.Set();

    LARGE_INTEGER time, frequency;
    QueryPerformanceCounter( &time );
    QueryPerformanceFrequency( &frequency );
    const float timef = float( (time.QuadPart-base_time.QuadPart) * 1000 / frequency.QuadPart );
    glRotatef( timef*45.0f/1000.0f, 0, -1, 0 );

    const GLint num_units = std::min( max_texture_units, GLint(texture.size()) );
    for ( GLint i=0; i < num_units; i++ )
        if ( texture[i] )
            glActiveTexture( GL_TEXTURE0 + i ),
            glBindTexture( GL_TEXTURE_2D, texture[i] );

    if ( texenv_setup.empty() )
        glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
    else
        SetTexEnv( texenv_setup );

    if ( IsDlgButtonChecked( ID_ALPHABLEND ) == BST_CHECKED )
        glEnable( GL_BLEND ),
        glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
    else
        glDisable( GL_BLEND );

    struct VERTEX {
        float u, v;
        float nx, ny, nz;
        float x, y, z;
    };
    VERTEX vertices[] = {
        { 0, 0, 0, 0, 1,-1,-1, 0 },
        { 1, 0, 0, 0, 1, 1,-1, 0 },
        { 0, 1, 0, 0, 1,-1, 1, 0 },

        { 1, 1, 0, 0, 1, 1, 1, 0 },
        { 0, 1, 0, 0, 1,-1, 1, 0 },
        { 1, 0, 0, 0, 1, 1,-1, 0 },

        { 0, 0, 0, 0,-1,-1,-1, 0 },
        { 0, 1, 0, 0,-1,-1, 1, 0 },
        { 1, 0, 0, 0,-1, 1,-1, 0 },

        { 1, 1, 0, 0,-1, 1, 1, 0 },
        { 1, 0, 0, 0,-1, 1,-1, 0 },
        { 0, 1, 0, 0,-1,-1, 1, 0 },
    };
/*
    // Set vertex array pointers
    glEnableClientState( GL_VERTEX_ARRAY );
    glVertexPointer( 3, GL_FLOAT, sizeof(VERTEX), &vertices[0].x );
    glEnableClientState( GL_NORMAL_ARRAY );
    glNormalPointer( GL_FLOAT, sizeof(VERTEX), &vertices[0].nx );
    for ( GLint i=0; i < num_units; i++ ) {
        glClientActiveTexture( GL_TEXTURE0 + i );
        if ( texture[i] )
            glEnableClientState( GL_TEXTURE_COORD_ARRAY ),
            glTexCoordPointer( 2, GL_FLOAT, sizeof(VERTEX), &vertices[0].u );
        else
            glDisableClientState( GL_TEXTURE_COORD_ARRAY );
    }

    // Draw mesh
    glDrawArrays( GL_TRIANGLES, 0, 12 );

    // Deactivate vertex arrays
    glDisableClientState( GL_VERTEX_ARRAY );
    glDisableClientState( GL_NORMAL_ARRAY );
    for ( GLint i=0; i < num_units; i++ )
        glClientActiveTexture( GL_TEXTURE0 + i ),
        glDisableClientState( GL_TEXTURE_COORD_ARRAY );
*/
/*
    glInterleavedArrays( GL_T2F_N3F_V3F, 0, vertices );
    glDrawArrays( GL_TRIANGLES, 0, 12 );
*/

    glBegin( GL_TRIANGLES );
        for ( int i=0; i < 12; i++ ) {
            const VERTEX& v = vertices[i];
            //glTexCoord2f( v.u, v.v );
            for ( GLint k=0; k < num_units; k++ )
                glMultiTexCoord2f( GL_TEXTURE0 + k, v.u, v.v );
            glNormal3f( v.nx, v.ny, v.nz );
            glVertex3f( v.x, v.y, v.z );
        }
    glEnd();

    DefaultTexEnv();
    glDisable( GL_BLEND );

    glFlush();
    SwapBuffers( hdc );
}

GLint MainWindow::LoadTexture( const std::string& filename )
{
    File file( filename.c_str() );
    if ( ! file.IsOpen() )
        return 0;
    const int size = file.GetSize();
    vector<char> buffer( size );
    if ( file.Read( &buffer[0], size ) != size )
        return 0;

    if ( ! PNGDecoder::CanDecode( buffer ) )
        return 0;
    vector<char> image;
    int width, height;
    PNGDecoder::DecodeImage( buffer, &image, &width, &height );

    vector<char> tmp( width*4 );
    for ( int i=0; i < height>>1; i++ ) {
        const int line1 = i * width * 4;
        const int line2 = (height - 1 - i) * width * 4;
        memcpy( &tmp[0], &image[line1], width*4 );
        memcpy( &image[line1], &image[line2], width*4 );
        memcpy( &image[line2], &tmp[0], width*4 );
    }

    GLuint texture = 0;
    glGenTextures( 1, &texture );
    glBindTexture( GL_TEXTURE_2D, texture );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
//  glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, &image[0] );
    gluBuild2DMipmaps( GL_TEXTURE_2D, GL_RGBA, width, height, GL_RGBA, GL_UNSIGNED_BYTE, &image[0] );

#define GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT 0x84FF
#define GL_TEXTURE_MAX_ANISOTROPY_EXT 0x84FE
    if ( strstr( (const char*)glGetString( GL_EXTENSIONS ), "GL_EXT_texture_filter_anisotropic" ) ) {
        GLfloat max_anisotropy = 1;
        glGetFloatv( GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, &max_anisotropy );
        glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, max_anisotropy );
    }

    return texture;
}

bool MainWindow::SetTexEnv( const std::vector<TEXENVVALUE> setup )
{
    if ( ! glActiveTexture ) return false;
    glGetError();
    for ( size_t i=0; i < setup.size(); i++ ) {
        const TEXENVVALUE& param = setup[i];
        switch ( param.param ) {

        case GL_ACTIVE_TEXTURE:
            glActiveTexture( GL_TEXTURE0 + param.valuei );
            glEnable( GL_TEXTURE_2D );
            break;

        case UNITDISABLE:
            glActiveTexture( GL_TEXTURE0 + param.valuei );
            glDisable( GL_TEXTURE_2D );
            break;

        case GL_TEXTURE_ENV_COLOR:
            glTexEnvfv( GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, param.valuef );
            break;

        default:
            glTexEnvi( GL_TEXTURE_ENV, param.param, param.valuei );
            break;

        }
        if ( glGetError() != GL_NO_ERROR )
            return false;
    }
    return true;
}

void MainWindow::DefaultTexEnv()
{
    if ( ! glActiveTexture ) return;
    for ( GLint i=1; i < max_texture_units; i++ ) {
        glActiveTexture( GL_TEXTURE0 + i );
        glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
        glDisable( GL_TEXTURE_2D );
    }
    glActiveTexture( GL_TEXTURE0 );
    glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
    glDisable( GL_TEXTURE_2D );
}

// vim:ts=4:sts=4:et

